# Set function paramters from gui window
# Derek Fujimoto
# April 2019

from tkinter import *
from tkinter import ttk
from bfit import logger_name
import matplotlib.pyplot as plt
import logging
import bdata as bd
from bfit.fitting.FunctionPlacer import FunctionPlacer

import numpy as np

# ========================================================================== #
class gui_param_popup(object):
    """
        Popup window for graphically finding input parameters. 
        
        data:           bdata object 
        fig:            maplotlib figure object
        fitter:         fit_tab.fitter obje (defined in default_routines.py)
        first:          if True, first time through fixing parameters
        fname:          name of the function 
        logger:         logging variable
        mode:           1 or 2 to switch between run modes
        n_components:   number of components in the fit function
        parnames:       tuple of names of the parameters
        p0:             dictionary of StringVar objects to link parameters
        selection:      StringVar, track run selection
        win:            TopLevel window
        xy:             (x,asym,dasym) tuple
    """

    # parameter mapping
    parmap = {  '1/T1':'lam',
                'amp':'amp',
                'beta':'beta',
                'baseline':'base',
                'peak':'peak',
                'width':'width',
                'height':'amp',
                'sigma':'width',
                'mean':'peak',
             }

    # ====================================================================== #
    def __init__(self,bfit):
        self.bfit = bfit
        self.first = True   # default
        
        # get logger
        self.logger = logging.getLogger(logger_name)
        self.logger.info('Initializing gui param popup')
        
        # make a new window
        self.win = Toplevel(bfit.mainframe)
        self.win.title('Find P0')
        frame = ttk.Frame(self.win,relief='sunken',pad=5)
        
        # Labels
        ttk.Label(frame,text="Select Run").grid(column=0,row=0,sticky=E)
        
        # box for run select
        self.selection = StringVar()
        select_box = ttk.Combobox(frame,textvariable=self.selection,
                                  state='readonly')
        select_box.bind('<<ComboboxSelected>>',self.setup)
        
        # get run list
        runlist = list(self.bfit.fit_files.fit_lines.keys())
        runlist.sort()
        # ~ runlist = ['('+r.split('.')[0]+') '+r.split('.')[1] for r in runlist]
        select_box['values'] = runlist
        
        # gridding
        frame.grid(column=0,row=1,sticky=(N,W,E,S))
        select_box.grid(column=0,row=1,sticky=E)
                
    # ====================================================================== #
    def setup(self,*args):
        """Get parameters for placing function and start the run squence"""
        
        # get run selection 
        run_id = self.selection.get()
        self.logger.info('Running P0 GUI finder on run %s',run_id)
        
        # get data
        self.data = self.bfit.data[run_id]
        mode = self.data.mode
        
        # mode switching
        if mode in ('20','2h'): self.mode = 2
        elif mode in ('1f',):   self.mode = 1
        else:
            self.logger.warning('P0 Finder not configured for run mode %s',mode)
            print('P0 Finder not configured for run mode %s'%mode)
        
        # make new window 
        self.fig = plt.figure()
        ax = self.fig.add_subplot(111)
        
        # draw data
        omit = self.data.omit.get()
        if omit == self.bfit.fetch_files.bin_remove_starter_line:
            omit = ''
        self.xy = self.data.asym('c',rebin=self.data.rebin.get(),omit=omit)
        ax.errorbar(*self.xy,fmt='.')
        
        # plot elements - don't do tight_layout here - blocks matplotlib signals
        ax.set_ylabel('Asymmetry')
        if self.mode == 2:     ax.set_xlabel('Time (s)')
        elif self.mode == 1:   ax.set_xlabel('Frequency (MHz)')
        self.fig.show()
        
        # get parameters list and run finder 
        fit_tab = self.bfit.fit_files
        self.fitter = fit_tab.fitter
        self.n_components = fit_tab.n_component.get()
        self.fname = fit_tab.fit_function_title.get()
        self.parnames = self.fitter.gen_param_names(fn_name=self.fname,
                                          ncomp=self.n_components)
        parentry = fit_tab.fit_lines[run_id].parentry
        
        # make initial paramter list
        self.p0 = {k:parentry[k]['p0'][0] for k in parentry.keys()}
        
        self.run()
        
    # ====================================================================== #
    def run(self,comp=0):
        """
            Run the function placer
            
            comp = component number to run
        """
        
        try:
            del self.fplace
        except AttributeError:
            pass
        else:
            self.fig.axes[0].cla()
            self.fig.axes[0].errorbar(*self.xy,fmt='.')
        
        # end condition, or start the cycle again
        if comp >= self.n_components:
            if self.n_components > 1: 
                self.first = False
                comp = 0
            else: 
                return
        
        # get the number of components in the line (after first round, keep all)
        if self.first:  ncomp = comp+1
        else:           ncomp = self.n_components
        
        # ensure matplotlib signals work. Not sure why this is needed.
        self.fig.tight_layout()
    
        # get parameter names and fitting functions - multi component
        if self.n_components > 1:
            
            # get parameter names of all previously set components 
            parnames_prev = []
            for c in range(ncomp-1 if self.first else ncomp):
                if c == comp: continue
                parnames_prev.extend([p for p in self.parnames if str(c) in p])
            
            # get parameter names of components to set now
            parnames_now = [p for p in self.parnames if str(comp) in p]
            
            # baseline name 
            if comp == 0 and self.first:    parnames_now.append('baseline')
            else:                           parnames_prev.append('baseline')
            
            # translate keynames: input to original
            parnames_now_conv = {self.parmap[k.split('_')[0]]:k for k in parnames_now}
            
            # static values 
            p = {p:float(self.p0[p].get()) for p in parnames_prev}  
            
            # p0 from current iteration
            p0 = {self.parmap[p.split('_')[0]]:self.p0[p] for p in parnames_now}
            
            # get fitting function 
            if self.mode == 1:
                f1 = self.fitter.get_fn(self.fname,ncomp=ncomp)
                
                # make decorator for fitting function
                if comp == 0 and self.first:
                    def fn(x,peak,width,amp,base):
                        
                        # add to input dictionary
                        p[parnames_now_conv['peak']] = peak
                        p[parnames_now_conv['width']] = width
                        p[parnames_now_conv['amp']] = amp
                        p[parnames_now_conv['base']] = base
                        
                        # get the order right
                        p_in = [p[k] for k in self.parnames if k in p.keys()]
                        return f1(x,*p_in)
                else:
                    def fn(x,peak,width,amp):
                        
                        # add to input dictionary
                        p[parnames_now_conv['peak']] = peak
                        p[parnames_now_conv['width']] = width
                        p[parnames_now_conv['amp']] = amp
                        
                        # get the order right
                        p_in = [p[k] for k in self.parnames if k in p.keys()]
                        return f1(x,*p_in)
               
            elif self.mode == 2:
                f1 = self.fitter.get_fn(self.fname,ncomp=ncomp,
                                           pulse_len=self.data.get_pulse_s(),
                                           lifetime=bd.life.Li8)
                                           
                # make decorator for fitting function
                if comp == 0 and self.first:
                    
                    if 'beta' in parnames_now_conv.keys():
                        def fn(x,lam,amp,beta,base):
                            
                            # add to input dictionary
                            p[parnames_now_conv['lam']] = lam
                            p[parnames_now_conv['amp']] = amp
                            p[parnames_now_conv['beta']] = beta
                            p[parnames_now_conv['base']] = base
                            
                            # get the order right
                            p_in = [p[k] for k in self.parnames if k in p.keys()]
                            return f1(x,*p_in)
                    else:
                        def fn(x,lam,amp,base):
                            
                            # add to input dictionary
                            p[parnames_now_conv['lam']] = lam
                            p[parnames_now_conv['amp']] = amp
                            p[parnames_now_conv['base']] = base
                            
                            # get the order right
                            p_in = [p[k] for k in self.parnames if k in p.keys()]
                            return f1(x,*p_in)               
                            
                else:
                    
                    if 'beta' in parnames_now_conv.keys():
                        def fn(x,lam,amp,beta):
                            
                            # add to input dictionary
                            p[parnames_now_conv['lam']] = lam
                            p[parnames_now_conv['amp']] = amp
                            p[parnames_now_conv['beta']] = beta
                            
                            # get the order right
                            p_in = [p[k] for k in self.parnames if k in p.keys()]
                            return f1(x,*p_in)
                    else:
                        def fn(x,lam,amp):
                            
                            # add to input dictionary
                            p[parnames_now_conv['lam']] = lam
                            p[parnames_now_conv['amp']] = amp
                            
                            # get the order right
                            p_in = [p[k] for k in self.parnames if k in p.keys()]
                            return f1(x,*p_in)               
                            
                                    
        # get parameter names and fitting functions - single component
        else:
            
            # get paramters, translating the names
            p0 = {self.parmap[k]:self.p0[k] for k in self.p0.keys()}
    
            # get fitting function 
            if self.mode == 1:
                f1 = self.fitter.get_fn(self.fname,ncomp=ncomp)
                fn = lambda x,peak,width,amp,base : f1(x,peak,width,amp,base)
                    
            elif self.mode == 2:
                f1 = self.fitter.get_fn(self.fname,ncomp=ncomp,
                                           pulse_len=self.data.get_pulse_s(),
                                           lifetime=bd.life.Li8)
                
                if 'beta' in self.parnames:
                    fn = lambda x,lam,amp,beta,base : f1(x,lam,amp,beta,base)
                else:
                    fn = lambda x,lam,amp,base : f1(x,lam,amp,base)
            
        # start recursive function placement
        self.fig.canvas.mpl_connect('close_event',self.cancel)
        self.fplace = FunctionPlacer(fig=self.fig,
                                     data=self.data,
                                     fn=fn,
                                     p0=p0,
                                     endfn=lambda:self.run(comp+1),
                                     base = float(self.p0['baseline'].get()))
        
    # ====================================================================== #
    def cancel(self,*args):
        if hasattr(self,'fplace'):  del self.fplace
        if hasattr(self,'fig'):     del self.fig
        
        self.win.destroy()






# Put a function on the active plot and adjust it's parameters with mouse buttons
# Derek Fujimoto
# April 2019

import bdata as bd
import matplotlib.pyplot as plt
import numpy as np
from bfit.fitting.functions import lorentzian # freq, peak, width, amp
from bfit.fitting.functions import pulsed_exp # time, lambda_s, amp
from bfit.fitting.functions import pulsed_strexp # time, lambda_s, beta, amp

class FunctionPlacer(object):
    
    npts = 500  # number of points used to draw line
    tau = bd.life.Li8   # lifetime of probe 
    
    # ======================================================================= #
    def __init__(self,fig,data,fn,p0,endfn,base=0):
        """
            fig:    pointer to matplotlib figure object to draw in
            data:   bdata object 
            fn:     function to draw and get the parameters for
            p0:     dictionary of StringVar corresponding to input parameters
            endfn:  function pointer to function to call at end of sequence. 
                        Called with no inputs
            base:   value of the baseline when we're not altering it
        
            fn needs input parameters with keys: 
            
                1F
                    peak, width, amp, base (optional)
                20/2H
                    amp, lam, beta (optional), base (optional)
        """
        # save input
        self.fig = fig
        self.fn = fn
        self.p0_variable = p0
        self.p0 = {k:float(p0[k].get()) for k in p0.keys()}
        self.endfn = endfn
        self.base = base
        
        self.mode = data.mode
        x = data.asym('c')[0]
        self.x = np.linspace(min(x),max(x),self.npts)
        
        # get axes for drawing
        self.ax = fig.axes
        if len(self.ax) == 0:
            self.ax = fig.add_subplot(111)
        else:
            self.ax = self.ax[0]
        
        # get ylims
        ylims = self.ax.get_ylim()
        
        # draw line with initial parameters
        self.line = self.ax.plot(self.x,fn(self.x,**self.p0),zorder=20)[0]
        
        # reset ylims
        self.ax.set_ylim(ylims)
        
        # start step
        self.step = 0
        
        # inital connect
        if self.mode in ('20','2h'):    self.connect_motion_20()
        elif self.mode in ('1f',):      self.connect_motion_1f()
        self.connect_click()
    
    # ======================================================================= #
    def connect_click(self):
        """Connect the mouse release events"""
        
        if self.mode in ('1f',):
            self.cidrelease = self.line.figure.canvas.mpl_connect(
                'button_release_event',self.connect_motion_1f)
        elif self.mode in ('20','2h'):
            self.cidrelease = self.line.figure.canvas.mpl_connect(
                'button_release_event',self.connect_motion_20)
        else:
            pass
            
    # ======================================================================= #
    def connect_motion_20(self,*args):
        """Connect the needed mouse motion events"""
        
        if self.step == 0:
            
            titlestr = ''
            
            # do baseline, if it exists    
            if 'base' in self.p0.keys():
                self.cidmotion_y = self.line.figure.canvas.mpl_connect(
                    'motion_notify_event', self.on_motion_20base)
                titlestr += 'base (ymove), '
                    
            # do T1
            self.cidmotion_x = self.line.figure.canvas.mpl_connect(
                'motion_notify_event', self.on_motion_20lam)
            titlestr += '1/T1 (xmove)'
            
            # do initial asymmetry
            self.cidscroll = self.line.figure.canvas.mpl_connect(
                    'scroll_event', self.on_scroll_20amp)
            titlestr += ', amp (scroll)'
            
            # do beta
            if 'beta' in self.p0.keys():
                self.cidkey = self.line.figure.canvas.mpl_connect(
                    'key_press_event', self.on_key_20beta)
                titlestr += ', beta (arrow keys)'
            
            self.ax.set_title(titlestr,fontsize='small')
            
        # END: disconnect
        else:
            
            if hasattr(self,'cidmotion_x'):   
                self.line.figure.canvas.mpl_disconnect(self.cidmotion_x)
            if hasattr(self,'cidmotion_y'):   
                self.line.figure.canvas.mpl_disconnect(self.cidmotion_y)
            if hasattr(self,'cidscroll'):   
                self.line.figure.canvas.mpl_disconnect(self.cidscroll)
            if hasattr(self,'cidkey'):   
                self.line.figure.canvas.mpl_disconnect(self.cidscroll)
            self.line.figure.canvas.mpl_disconnect(self.cidrelease)
            
            self.ax.set_title('')
            
            # display new parameters
            for k in self.p0.keys():
                self.p0_variable[k].set(str(self.p0[k]))
                
            # end the sequence
            self.endfn()
            
        self.step += 1
        self.fig.show()
    
    # ======================================================================= #
    def connect_motion_1f(self,*args):
        """Connect the needed mouse motion events"""
        
        # connect motion to setting the peak
        if self.step == 0:

            self.ax.set_title('Click to set peak position (x and y)',fontsize='small')
            self.cidmotion = self.line.figure.canvas.mpl_connect(
                'motion_notify_event', self.on_motion_1fpeak)
        
        # connect motion to setting the base
        elif self.step == 1:
            
            if 'base' in self.p0.keys():
                self.ax.set_title('Click to set baseline',fontsize='small')
                self.line.figure.canvas.mpl_disconnect(self.cidmotion)
                self.cidmotion = self.line.figure.canvas.mpl_connect(
                    'motion_notify_event', self.on_motion_1fbase)
            else:
                self.step += 1
                self.connect_motion_1f()
                
         # connect motion to setting the width
        elif self.step == 2:
            
            self.ax.set_title('Click to set width',fontsize='small')
            
            if hasattr(self,'cidmotion'):
                self.line.figure.canvas.mpl_disconnect(self.cidmotion)
            self.cidmotion = self.line.figure.canvas.mpl_connect(
                'motion_notify_event', self.on_motion_1fwidth)
            
        # END: disconnect
        else:
            self.ax.set_title('')
            self.line.figure.canvas.mpl_disconnect(self.cidmotion)
            self.line.figure.canvas.mpl_disconnect(self.cidrelease)
            
            # display new parameters
            for k in self.p0.keys():
                self.p0_variable[k].set(str(self.p0[k]))
            
            # end the sequence
            self.endfn()
        
        self.step += 1
        self.fig.show()
    
    # ======================================================================= #
    def on_motion_20base(self,event):
        """Updated the baseline on mouse movement"""
        
        # check event data
        if event.ydata is not None:
            
            self.p0['base'] = event.ydata
            self.line.set_ydata(self.fn(self.x,**self.p0))
            self.line.figure.canvas.draw()
     
    # ======================================================================= #
    def on_scroll_20amp(self,event):
        """Updated the initial asymmetry on mouse movement"""
     
        # check event data
        if event.step is not None:
            
            self.p0['amp'] += self.p0['amp']*0.05*event.step
            self.line.set_ydata(self.fn(self.x,**self.p0))
            self.line.figure.canvas.draw()
 
    # ======================================================================= #
    def on_key_20beta(self,event):
        """Updated the initial asymmetry on mouse movement"""
     
        beta = self.p0['beta']
     
        # check event data
        if event.key in ('up','right'):
            beta += beta*0.05
            
        elif event.key in ('down','left'):
            beta -= beta*0.05            
        else:
             return

        self.p0['beta'] = min(max(beta,0),1)
        self.line.set_ydata(self.fn(self.x,**self.p0))
        self.line.figure.canvas.draw()
        
    # ======================================================================= #
    def on_motion_20lam(self,event):
        """Updated 1/T1 on mouse movement"""
        
        # set lambda
        if event.xdata is not None:
            self.p0['lam'] = max(0,1/(event.xdata*2))
            self.line.set_ydata(self.fn(self.x,**self.p0))
            self.line.figure.canvas.draw()
                
    # ======================================================================= #
    def on_motion_1fpeak(self,event):
        """Updated the peak position on mouse movement"""
    
        # check event data
        if event.xdata is not None and event.ydata is not None:
            
            self.p0['peak'] = event.xdata
            
            if 'base' in self.p0.keys():
                self.p0['base'] = event.ydata+self.p0['amp']
            else:
                self.p0['amp'] = self.base-event.ydata
            
            self.line.set_ydata(self.fn(self.x,**self.p0))
            self.line.figure.canvas.draw()
        
    # ======================================================================= #
    def on_motion_1fbase(self,event):
        """Updated the baseline position on mouse movement"""
        
        # check event data
        if event.xdata is not None and event.ydata is not None:
            
            self.p0['amp'] -= self.p0['base']-event.ydata
            self.p0['base'] = event.ydata
            
            self.line.set_ydata(self.fn(self.x,**self.p0))
            self.line.figure.canvas.draw()
    
    # ======================================================================= #
    def on_motion_1fwidth(self,event):
        """Updated the width on mouse movement"""
        
        # check event data
        if event.xdata is not None and event.ydata is not None:
            
            self.p0['width'] = abs(self.p0['peak']-event.xdata)
            self.line.set_ydata(self.fn(self.x,**self.p0))
            self.line.figure.canvas.draw()
